# Internet Of Things - Worksheet 2 part 2

![image](src/uwe.png)

**Module Name:** Internet Of Things

**Module Code:** UFCFVK‑15‑2

**Module Leader Name:** Benedict R. Gaster

- [Internet Of Things - Worksheet 2 part 2](#internet-of-things---worksheet-2-part-2)
  - [Description](#description)
    - [Task 1: Binary heap.](#task-1-binary-heap)
    - [Task 2: Python encode / decode.](#task-2-python-encode--decode)
    - [Task 3: Server echo.](#task-3-server-echo)
  - [Demos](#demos)
  - [Getting Started](#getting-started)
    - [Dependencies](#dependencies)
    - [Installing](#installing)
  - [Authors](#authors)
  - [Version History](#version-history)
  - [License](#license)

## Description
Worksheet(2 part 2)  
-This worksheet explores using servers on the csct cloud and communicating via ham standards  
**This is a rep for only worksheet 2 part 2.**
* [IOT - Worksheet 2 part 1 ](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1) can be found here  
* [IOT - Worksheet 2 part 2 ](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-2)  **you are here**
* [IOT - Worksheet 3](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-3) can be found here


### Task 1: Binary heap.
* A binary heap is a data structer that takes the form of a binary tree. In our case this will be a string of character's stored in breadth first order. Following a set of rules and creating an algorithm around those rules we can encode and decode the binary tree.
* ***inserter.py*** is from the pervious worksheet which can be found here: [IOT - Worksheet 2 part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1)  is designed to take an heap string and add values to that heap. It also will add new layers by adding the correct amount of Null characters. This is what the tree looks like in heap form now.![image](src/screenshot1.PNG). The function to insert is very simple and the next worksheet builds on a similar method so it will be very useful. ![image](src/screenshot2.PNG) When we run our new characters to be inserted ![image](src/screenshot3.PNG) we get ![image](src/screenshot4.PNG). This is useful for this task as we will need this string outputted from this for our `decode_bt()`
* ***def decode_bt(string)*** is located in **binaryheap.py** similar to the last approach in [part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1) will will be using a binary tree to decode, however, as discussed this will be in a heap. The heap string is stored in a known format so if we follow that we can decode. Using `tree[0]` we can access the root element of the tree. Using `tree[(2*p)]` where p is the parent\current node we can access the left child. For right child we can use `tree[(2*p)+1]` where p is the parent\current node. Using this logic we can now begin to decode the heap. ![image](src/screenshot5.PNG)
* ***unittest.py*** unittest is an inbuilt Python library. Its a test runner and is specially design for running tests and debugging. This one contains a few useful tools that we are going to use. It works similarly to the assert function. We've written a few test cases to test things like: Correct, incorrect, invalid inputs for both encode and decode. After running the unittests we spotted a mistake where the decode outputted `-` instead of a `space character`  ![image](src/screenshot6.PNG) to fix this we replaced the head of the heap with a space to return a space character when there is a space.![image](src/screenshot7.PNG) ![image](src/screenshot8.PNG)
*  ***The differences***
   *  The different approaches may not seem to to matter, however, on a large scale program the smallest of differences can have large impacts when scaled up. Here is a few comparisions on the decoder developed here and the OOP decoder from [part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1)
   *  **Advantages of OOP implementation**
      *  The **node()** class has many methods. These methods can change values of the tree meaning the tree can be easily altered at any point in the program by calling the **def setLeftChildNode(nodes)** and **def setLeftChildNode(nodes)**. However, with the heap implementation with would need to use something similar to `convert.py` replacing the string.
   *  **Disadvantages of OOP implementation**
      *  If we run `runtime.py` which is a program to test the run time of the programs we can see that the run time of the heap is **10x** faster than OOP, why is this? Current each time **decode** runs it creates the tree object. This takes time to set all the children however  the heap already is setup. If we were to test this where both are already test up we would likely see similar times as they are both **0(n)** ![image](src/screenshot9.PNG) ![image](src/screenshot10.PNG)
* ***A dictionary implementation?***
  * Using dictionaries would result in a hybrid implementation of the two. Where dictionaries contain the child's dictionaries. This would likely be done using json format. This would be fairly robust and less costly to later children. However, dictionaries are costly and slow. [An example: p8](http://www.cs.cmu.edu/~15110-f19/slides/week8-2-trees.pdf)
___
### Task 2: Python encode / decode.
* Task two builds on [part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1) and explores communication via Ham radios standards.
* *The CSCT cloud server has a port open on 10105 which contains a morse encoder and decoder. We are required to connect to the service and test its functionality*.
* To connect to the port on the csct server we need to be able to access the port. Todo this we can port ford the port via VS Code. Using the command `Ctrl + Shift + P`  we can bring up the command wind and open ports: ![image](src/screenshot11.PNG) if we then add the port 10105 ![image](src/screenshot12.PNG) should be able to access that port on our local machine using a browser to access `localhost:10101` as we can see it has worked and we can now connect to the 10105 port on the csct cloud server: ![image](src/screenshot13.PNG)
* Here are a few example runs of testing the encoder / decoder service:
  * ![image](src/screenshot13.PNG) ***Sender, Receiver, hello*** encodes to ***..-. . -.-. .. . ...- . .-. -.. . ... . -. -.. . .-. -...- .... . .-.. .-.. --- -...- -.--.***
* Task two requires use to create a python morse encoder / decoder using a binary tree.
* Morse.py contains two functions **encode()** and **decode()**. They do what you might expect. encode() takes ascii string and returns a string of `_` or `.` for the corresponding letter. This is done with a binary tree implementation.
* ***The ham encoder/decoder:***
  * Ham follows a basic format: `receiver 'de' sender msg` where **=** is place for the message and **=(** after.
  * ***def encode_ham(sender, receiver, msg)*** takes the three arguments and formats them to ham stands. After doing that it will then run the new string through the morse encoder from [part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1) and returns the output. ![image](src/screenshot14.PNG)
* ***The ham decoder:***
  * Ham follows a basic format: `receiver 'de' sender msg` where **=** is place for the message and   **=(** after. We can use this to decode incoming messages.
  * ***def decode_ham(msg)*** takes one argument of a morse string and returns a python tuple of the **sender receiver** and **message**. Again using the morse decoder from [part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1). ![image](src/screenshot15.PNG)
* ***unittest.py*** unittest is an inbuilt Python library. Its a test runner and is specially design for running tests and debugging. This one contains a few useful tools that we are going to use. It works similarly to the assert function. We've written a few test cases to test things like: Correct, incorrect, invalid inputs for both encode and decode. After running the unittests I found an error. ![image](src/screenshot16.PNG) The command `ret.split("DE")` is used to split the **sender** from the **receiver** however it will keep splitting even after the first DE. Simply adding a limit will solve this problem. ![image](src/screenshot17.PNG) ![image](src/screenshot18.PNG)
___
### Task 3: Server echo.
* *The CSCT cloud server has a port open on 10102 which contains two services, **echo**, sends back the message to the sender and **time**, sends back the time.
* To communicate with the server we simple need to bring a few things together from previous tasks.
* ***server.py*** connects to the servers websocket and sends and receives json from the server. It packages up the necessary components, type, client_id, and payload in json to be able to comminate with the protocol on the server. ![image](src/screenshot19.PNG)
* ***hamradio.py*** Please see **The ham encoder/decoder:** section to see how this works.
* ***hamserver.py*** This brings the two files above into one file to be able to encode and decode the server messages as it operates using the ham format. ![image](src/screenshot20.PNG).
* ***unittest.py*** unittest is an inbuilt Python library. Its a test runner and is specially design for running tests and debugging. This one contains a few useful tools that we are going to use. It works similarly to the assert function. We've written a few test cases to test things like: Correct, incorrect, invalid inputs for both encode and decode. Again after testing I found a **FUNDAMENTAL ERROR** I was using two spaces to indicate a separation of works when in fact it should be three. Heres how I went about fixing it. ![image](src/screenshot21.PNG) Here you can see that double spacing has appeared in the result this is unexpected. [part 1](https://gitlab.uwe.ac.uk/jo2-holdsworth/iot-worksheet-2-part-1)'s logic will need to be changed in both the **encode** and **decode** functions to accommodate the change.![image](src/screenshot22.PNG)![image](src/screenshot23.PNG)![image](src/screenshot25.PNG)![image](src/screenshot24.PNG). To be able to test the time function we must first get the time. `datetime` can do that for us. This method of testing is not bullet proof but it does run successfully most of the time on quick connections. ![image](src/screenshot26.PNG)
___
## Demos
* ***Task 1:***  
  <iframe width="560" height="315" src="https://www.youtube.com/embed/coAidIG8ZoQ" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share" allowfullscreen></iframe>
* ***Task 2:***  
  <iframe width="560" height="315" src="https://www.youtube.com/embed/hqghPQFcW4U" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share" allowfullscreen></iframe>
* ***Task 3:***  
  <iframe width="560" height="315" src="https://www.youtube.com/embed/BPZv5ajf3TM" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share" allowfullscreen></iframe>
___
## Getting Started

* Running the task files
  * Task 1: `python3 binaryheap.py`
  * Task 2: `python3 hamradio.py`
  * Task 3: `python3 hamserver.py`

### Dependencies

* datetime
* asyncio
* unittest

### Installing

* Nothing to install

## Authors

Contributors names and contact info

ex. Jack Holdsworth @ [my website](https://holdsworth.dev)

## Version History
* 0.4
  * README
  * Unittesting
* 0.3
  * task 2 done
  * started task 3
* 0.2
  * most of task 2 done
  * README made
* 0.1
  * init.

## License
This project is licensed under the GPL License - see the LICENSE.md file for details
